// VncSharp - .NET VNC Client Library
// Copyright (C) 2004  David Humphrey, Chuck Borgh, Matt Cyr
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace VncSharp.Encodings
{
	/// <summary>
	/// Abstract class representing an Encoded Rectangle to be read, decoded, and drawn.
	/// </summary>
	public abstract class EncodedRectangle : IDesktopUpdater
	{
		protected RfbProtocol rfb;
		protected Rectangle rectangle;
		protected Framebuffer framebuffer;

		// TODO: could pixels not be initialized once to the maximum possible size (i.e., 
		// framebuffer-width * framebuffer-height) and then an offset used to allow 
		// only portions to be read/written?  This would save having to create and then
		// throw away so much memory with each call to EncodedRectangle.Decode().
		
		protected int[] pixels;				// Pixel values will always be integers to 
											// match local representation (i.e., 32bpp)

		private const byte ALPHA = 0xFF;  // Alpha will always be 255 in calculations below

		public EncodedRectangle(RfbProtocol rfb, Framebuffer framebuffer, Rectangle rectangle)
		{
			this.rfb = rfb;
			this.framebuffer = framebuffer;
			this.rectangle = rectangle;
		}

		/// <summary>
		/// Gets the rectangle that needs to be decoded and drawn.
		/// </summary>
		public Rectangle UpdateRectangle {
			get {
				return rectangle;
			}
		}

		/// <summary>
		/// Obtain all necessary information from VNC Host (i.e., read) in order to Draw the rectangle, and store in colours[].
		/// </summary>
		public abstract void Decode();
		
		/// <summary>
		/// After calling Decode() an EncodedRectangle can be drawn to a Bitmap, which is the local representation of the remote desktop.
		/// </summary>
		/// <param name="desktop">The image the represents the remote desktop. NOTE: this image will be altered.</param>
		public unsafe virtual void Draw(Bitmap desktop)
		{
			// Lock the bitmap's scan-lines in RAM so we can iterate over them using pointers and update the area
			// defined in rectangle.
			BitmapData bmpd = desktop.LockBits(new Rectangle(new Point(0,0), desktop.Size), ImageLockMode.ReadWrite, desktop.PixelFormat);

			try {
				// For speed I'm using pointers to manipulate the desktop bitmap, which is unsafe.
				// Get a pointer to the start of the bitmap in memory (IntPtr) and cast to a 
				// Byte pointer (need void* first) so desktop can be traversed as GDI+ 
				// colour values form.
				int* pInt = (int*)(void*)bmpd.Scan0; 

				// Move pointer to position in desktop bitmap where rectangle begins
				pInt += rectangle.Y * desktop.Width + rectangle.X;
				
				int offset = desktop.Width - rectangle.Width;
				int row = 0;
				
				for (int y = 0; y < rectangle.Height; ++y) {
					row = y * rectangle.Width;

					for (int x = 0; x < rectangle.Width; ++x) {
						*pInt++ = pixels[row + x];
					}

					// Move pointer to beginning of next row in rectangle
					pInt += offset;
				}
			} finally {
				desktop.UnlockBits(bmpd);
				bmpd = null;
			}		
		}

		/// <summary>
		/// Fills the given Rectangle with a solid colour (i.e., all pixels will have the same value--colour).
		/// </summary>
		/// <param name="rect">The rectangle to be filled.</param>
		/// <param name="colour">The colour to use when filling the rectangle.</param>
		protected void FillRectangle(Rectangle rect, int colour)
		{
			int ptr = 0;
			int offset = 0;

			// If the two rectangles don't match, then rect is contained within rectangle, and
			// ptr and offset need to be adjusted to position things at the proper starting point.
			if (rect != rectangle) {
				ptr = rect.Y * rectangle.Width + rect.X;	// move to the start of the rectangle in pixels
				offset = rectangle.Width - rect.Width;		// calculate the offset to get to the start of the next row
			}

			for (int y = 0; y < rect.Height; ++y) {
				for (int x = 0; x < rect.Width; ++x) {
					pixels[ptr++] = colour;	// colour every pixel the same
				}
				ptr += offset;	// advance to next row within pixels
			}
		}
		
		/// <summary>
		/// Fills the given Rectangle with pixel values read from the server (i.e., each pixel may have its own value).
		/// </summary>
		/// <param name="rect">The rectangle to be filled.</param>
		protected void FillRectangle(Rectangle rect)
		{
			int ptr = 0;
			int offset = 0;

			// If the two rectangles don't match, then rect is contained within rectangle, and
			// ptr and offset need to be adjusted to position things at the proper starting point.
			if (rect != rectangle) {
				ptr = rect.Y * rectangle.Width + rect.X;	// move to the start of the rectangle in pixels
				offset = rectangle.Width - rect.Width;		// calculate the offset to get to the start of the next row
			}

			for (int y = 0; y < rect.Height; ++y) {
				for (int x = 0; x < rect.Width; ++x) {
					pixels[ptr++] = GetGdiPlusPixel();	// every pixel needs to be read from server
				}
				ptr += offset;	// advance to next row within pixels
			}
		}

		protected int GetGdiPlusPixel()
		{
			// Read the appropriate number of bytes from the server based on the  width of pixels and conver 
			// to a GDI+ colour value (i.e., BGRA). NOTE: I had been doing this with a delegate that got set 
			// in the ctor and pointed to the appropriate function below.  However, this is faster--just not
			// as pretty :(
			switch (framebuffer.BitsPerPixel) {
				case 32:
					return Get32BitGdiPlusPixel();
				case 16:
					return Get16BitGdiPlusPixel();
				case 8:
					return Get8BitGdiPlusPixel();
				default:
					throw new ArgumentOutOfRangeException("BitsPerPixel", framebuffer.BitsPerPixel, "Valid VNC Pixel Widths are 8, 16, or 32 bits.");
			}
		}

		private int Get32BitGdiPlusPixel() 
		{
			// Read the pixel value
			byte[] b = rfb.ReadBytes(4);
			uint pixel = (uint)(b[0] & 0xFF | b[1] << 8 | b[2] << 16 | b[3] << 24);

			// Extract RGB intensities from pixel
			byte red   = (byte) ((pixel >> framebuffer.RedShift)   & framebuffer.RedMax);
			byte green = (byte) ((pixel >> framebuffer.GreenShift) & framebuffer.GreenMax);
			byte blue  = (byte) ((pixel >> framebuffer.BlueShift)  & framebuffer.BlueMax);

			return ToGdiPlusOrder(red, green, blue);			
		}
		
		private int Get16BitGdiPlusPixel() 
		{
			// Same story as Get32BitGdiPlusPixel, except with 16 bits instead of 32.
			// NOTE: the 16 bit pixel value uses a 565 layout (i.e., 5 bits
			// for Red, 6 for Green, 5 for Blue).  As such, I'm doing an extra
			// shift below for each colour to get from 565 to 888 in order to 
			// return a full 32-bit pixel value.
			byte[] b = rfb.ReadBytes(2);
			ushort pixel = (ushort)(b[0] & 0xFF | b[1] << 8);

			byte red   = (byte) (((pixel >> framebuffer.RedShift)   & framebuffer.RedMax)   << 3);  // 5 bits to 8
			byte green = (byte) (((pixel >> framebuffer.GreenShift) & framebuffer.GreenMax) << 2);  // 6 bits to 8
			byte blue  = (byte) (((pixel >> framebuffer.BlueShift)  & framebuffer.BlueMax)  << 3);  // 5 bits to 8

			return ToGdiPlusOrder(red, green, blue);			
		}
		
		private int Get8BitGdiPlusPixel() 
		{
			// TODO: 8-Bit colour maps are not implemented yet.  If you want to do it,
			// you'll need to read the colour map from the server, which will be 32-bit
			// colour values that need to be keyed by position in an array.  Then when
			// you read a byte, it will be the position of the colour to use in the map.
			byte pixel = rfb.ReadByte();
			throw new NotImplementedException("8 bpp colour maps are not currently supported.");
		}
		
		private int ToGdiPlusOrder(byte red, byte green, byte blue)
		{
			// Put colour values into proper order for GDI+ (i.e., BGRA)
			return (int) (blue & 0xFF | green << 8 | red << 16 | ALPHA << 24);			
		}
	}
}


// ------------------------------------------------------------------------------------
// NOTE: This code is unused, and is only here to remind me what to do if I
// If ever I want to convert this code to VB.NET, which doesn't support pointers.
// ------------------------------------------------------------------------------------
//		private static Color GetPixel(Point pos, System.Drawing.Imaging.BitmapData bmd)
//		{
////			if (CheckPixel(pos, bmd)) {
//			//always assumes 32 bit per pixels
//			int offset=pos.Y*bmd.Stride+(4*pos.X);
//			return Color.FromArgb(
//				Marshal.ReadByte(bmd.Scan0,offset+2),
//				Marshal.ReadByte(bmd.Scan0,offset+1),
//				Marshal.ReadByte(bmd.Scan0,offset));
////			} else {
////				return Color.FromArgb(0,0,0,0);
////			}
//		}
//		
//		private static void SetPixel(Point pos, System.Drawing.Imaging.BitmapData bmd, Color c)
//		{
////			if (CheckPixel(pos,bmd)) {
//				//always assumes 32 bit per pixels
//				int offset=pos.Y*bmd.Stride+(4*pos.X);
//				Marshal.WriteByte(bmd.Scan0,offset+2,c.R);
//				Marshal.WriteByte(bmd.Scan0,offset+1,c.G);
//				Marshal.WriteByte(bmd.Scan0,offset,c.B);
//				Marshal.WriteByte(bmd.Scan0,offset+3,255);
////			}
//		}