// VncSharp - .NET VNC Client Library
// Copyright (C) 2004  David Humphrey, Chuck Borgh, Matt Cyr
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Drawing;
using System.Threading;
using System.Reflection;
using System.Windows.Forms;
using System.ComponentModel;
using System.Drawing.Imaging;
using System.Drawing.Drawing2D;

using VncSharp.Encodings;

namespace VncSharp
{
	/// <summary>
	/// Event Handler delegate declaration used by events that signal successful connection with the server.
	/// </summary>
	public delegate void ConnectCompleteHandler(object sender, ConnectEventArgs e);
	
	/// <summary>
	/// When connecting to a VNC Host, a password will sometimes be required.  Therefore a password must be obtained from the user.  A default Password dialog box is included and will be used unless users of the control provide their own Authenticate delegate function for the task.  For example, this might pull a password from a configuration file of some type instead of prompting the user.
	/// </summary>
	public delegate string AuthenticateDelegate();
	
	[ToolboxBitmap(typeof(RemoteDesktop), "Resources.vncviewer.ico")]
	/// <summary>
	/// The RemoteDesktop control takes care of all the necessary RFB Protocol and GUI handling, including mouse and keyboard support, as well as requesting and processing screen updates from the remote VNC host.  Most users will choose to use the RemoteDesktop control alone and not use any of the other protocol classes directly.
	/// </summary>
	public class RemoteDesktop : ScrollableControl
	{
		[Description("Raised after a successful call to the Connect() method.")]
		/// <summary>
		/// Raised after a successful call to the Connect() method.  Includes information for updating the local display in ConnectEventArgs.
		/// </summary>
		public event ConnectCompleteHandler ConnectComplete;
		
		[Description("Raised when the VNC Host drops the connection.")]
		/// <summary>
		/// Raised when the VNC Host drops the connection.
		/// </summary>
		public event EventHandler			ConnectionLost;
		
		/// <summary>
		/// Points to a Function capable of obtaining a user's password.  By default this means using the PasswordDialog.GetPassword() function; however, users of RemoteDesktop can replace this with any function they like, so long as it matches the delegate type.
		/// </summary>
		public AuthenticateDelegate GetPassword;
		
		Bitmap desktop;						// Internal representation of remote image.
		Image  designModeDesktop;			// Used when painting control in VS.NET designer
		VncClient vnc;						// The Client object handling all protocol-level interaction
		int port;							// The port to connect to on remote host
		bool passwordPending;				// After Connect() is called, a password might be required.

		public RemoteDesktop() : base()
		{
			// Since this control will be updated constantly, and all graphics will be drawn by this class,
			// set the control's painting for best user-drawn performance.
			SetStyle(ControlStyles.AllPaintingInWmPaint | 
					 ControlStyles.UserPaint			|
					 ControlStyles.DoubleBuffer			|
					 ControlStyles.ResizeRedraw			|
					 ControlStyles.Opaque,				
					 true);
			
			// Change the cursor to the "vnc" custor--a see-through dot
			Cursor = new Cursor(GetType(), "Resources.vnccursor.cur");

			// Allow only a portion of the remote desktop to be shown
			AutoScroll = true;
			AutoScrollMinSize = new Size(608, 427); // just a default for Design graphic. Will get changed once connected.
			
			// Show a screenshot of a Windows desktop from the manifest and cache to be used when painting in design mode
			designModeDesktop = Image.FromStream(Assembly.GetAssembly(GetType()).GetManifestResourceStream("VncSharp.Resources.screenshot.png"));
			
			// Users of the control can choose to use their own Authentication GetPassword() method via the delegate above.  This is a default only.
			GetPassword = new AuthenticateDelegate(PasswordDialog.GetPassword);
			port = 5900;	// default VNC port

			// Users should NOT attempt to send a password until one is required.
			passwordPending = false;
		}
		
		[DefaultValue(5900)]
		[Description("The port number used by the VNC Host (typically 5900)")]
		/// <summary>
		/// The port number used by the VNC Host (typically 5900).
		/// </summary>
		public int VncPort {
			get { 
				return port; 
			}
			set { 
				// Ignore attempts to use invalid port numbers
				if (value < 1 | value > 65535) value = 5900;
				port = value;	
			}
		}
		
		// This is a hack to get around the issue of DesignMode returning
		// false when the control is being removed from a form at design time.
		// First check to see if the control is in DesignMode, then work up 
		// to also check any parent controls.  DesignMode returns False sometimes
		// when it is really True for the parent. Thanks to Claes Bergefall for the idea.
		protected new bool DesignMode {
			get {
				if (base.DesignMode) {
					return true;
				} else {
					Control parent = Parent;
					
					while (parent != null) {
						if (parent.Site != null && parent.Site.DesignMode) {
							return true;
						}
						parent = parent.Parent;
					}
					return false;
				}
			}
		}

		/// <summary>
		/// Returns a more appropriate default size for initial drawing of the control at design time
		/// </summary>
		protected override Size DefaultSize {
			get { 
				return new Size(400, 200); 
			}
		}

		// This event handler deals with Frambebuffer Updates coming from the host. An
		// EncodedRectangle object is passed via the VncEventArgs (actually an IDesktopUpdater
		// object so that *only* Draw() can be called here--Decode() is done elsewhere).
		// The VncClient object handles thread marshalling onto the UI thread.
		protected void VncUpdate(object sender, VncEventArgs e)
		{
			e.DesktopUpdater.Draw(desktop);
		
			// Before Invalidating the control, specify the rectangle the needs to be
			// re-painted.  This speeds OnPaint dramatically, since the rest of the 
			// screen doesn't need to change.  Also, adjust the invalidated rectangle
			// to accomodate scrolling or centering, since the desktop will either be 
			// clipped or centered on the screen.
			int x, y;
			
			if (ClientSize.Width > desktop.Size.Width) {
				int dx = (ClientRectangle.Width - desktop.Width) / 2;
				x = e.DesktopUpdater.UpdateRectangle.X + dx;
			} else {
				x = e.DesktopUpdater.UpdateRectangle.X + AutoScrollPosition.X;
			}

			if (ClientSize.Height > desktop.Size.Height ) {
				int dy = (ClientRectangle.Height - desktop.Height) / 2;
				y = e.DesktopUpdater.UpdateRectangle.Y + dy;
			} else {
				y = e.DesktopUpdater.UpdateRectangle.Y + AutoScrollPosition.Y;
			}

			Rectangle adjusted = new Rectangle(x,
											   y,
											   e.DesktopUpdater.UpdateRectangle.Width,
											   e.DesktopUpdater.UpdateRectangle.Height);			

			Invalidate(adjusted);
		}

		/// <summary>
		/// Connect to a VNC Host and determine whether or not the server requires a password.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		public void Connect(string host)
		{
			// Use Display 0 by default.
			Connect(host, 0);
		}

		/// <summary>
		/// Connect to a VNC Host and determine whether or not the server requires a password.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <param name="display">The Display number (used on Unix hosts).</param>
		public void Connect(string host, int display)
		{
			// TODO: Should this be done asynchronously so as not to block the UI?  Since an event 
			// indicates the end of the connection, maybe that would be a better design.

			if (host == null) throw new ArgumentNullException("host");
			if (display < 0) throw new ArgumentOutOfRangeException("display", display, "Display number must be a positive integer.");
			
			// Start protocol-level handling and determine whether a password is needed
			vnc = new VncClient();
			passwordPending = vnc.Connect(host, display, VncPort);
			
			if (passwordPending) {
				// Server needs a password, so call which ever method is refered to by the GetPassword delegate.
				string password = GetPassword();

				if (password == null) {
					// No password could be obtained (e.g., user clicked Cancel), so stop connecting
					Disconnect();
					OnConnectionLost(EventArgs.Empty);										
				} else {
					Authenticate(password);
				}
			} else { 
				// No password needed, so go ahead and Initialize here
				Initialize();
			}
		}

		/// <summary>
		/// Authenticate with the VNC Host using a user supplied password.
		/// </summary>
		/// <exception cref="System.InvalidOperationException">Thrown if the VNC Host does not require password authentication. This method should only be called after Connect() returns True.</exception>
		/// <exception cref="System.NullReferenceException">Thrown if the password is null.</exception>
		/// <param name="password">The user's password.</param>
		public void Authenticate(string password)
		{
			if (!passwordPending) throw new InvalidOperationException("Authentication is only required when Connect() returns True and the VNC Host requires a password.");
			if (password == null) throw new NullReferenceException("password");

			passwordPending = false;  // repeated calls to Authenticate should fail.
			if (vnc.Authenticate(password)) {
				Initialize();
			} else {		
				OnConnectionLost(EventArgs.Empty);										
			}	
		}

		/// <summary>
		/// After protocol-level initialization and connecting is complete, the local GUI objects have to be set-up, and requests for updates to the remote host begun.
		/// </summary>
		protected void Initialize()
		{
			// Finish protocol handshake with host now that authentication is done.
			vnc.Initialize();
			
			// Create a buffer on which updated rectangles will be drawn and draw a "please wait..." 
			// message on the buffer for initial display until we start getting rectangles
			SetupDesktop();
			
			// Update the scrollable region to reflect the size of the framebuffer
			AutoScrollMinSize = new Size(vnc.Framebuffer.Width, vnc.Framebuffer.Height);
	
			// Tell the user of this control the necessary info about the desktop in order to setup the display
			OnConnectComplete(new ConnectEventArgs(vnc.Framebuffer.Width,
												   vnc.Framebuffer.Height, 
												   vnc.Framebuffer.DesktopName));

			// Start getting updates from the remote host (vnc.StartUpdates will begin a worker thread).
			vnc.VncUpdate += new VncUpdateHandler(VncUpdate);
			vnc.StartUpdates();
		}

		/// <summary>
		/// Creates and initially sets-up the local bitmap that will represent the remote desktop image.
		/// </summary>
		protected void SetupDesktop()
		{
			// Create a new bitmap to cache locally the remote desktop image.  Use the geometry of the
			// remote framebuffer, and 32bpp pixel format (doesn't matter what the server is sending--8,16,
			// or 32--we always draw 32bpp here for efficiency).
			desktop = new Bitmap(vnc.Framebuffer.Width, vnc.Framebuffer.Height, PixelFormat.Format32bppPArgb);
			
			// Draw a "please wait..." message on the local desktop until the first
			// rectangle(s) arrive and overwrite with the desktop image.
			using (Graphics g = Graphics.FromImage(desktop)) {
				g.FillRectangle(Brushes.Black, vnc.Framebuffer.Rectangle);

				StringFormat format = new StringFormat();
				format.Alignment = StringAlignment.Center;
				format.LineAlignment = StringAlignment.Center;

				g.DrawString("Please Wait, Connecting to VNC Host...", 
							 new Font("Arial", 12), 
							 new SolidBrush(Color.White), 
							 new PointF(vnc.Framebuffer.Width / 2, vnc.Framebuffer.Height / 2), format);
			}
		}
		
		/// <summary>
		/// Stops the remote host from sending further updates and disconnects.
		/// </summary>
		public void Disconnect()
		{
			vnc.StopUpdates();
		}

		protected override void Dispose(bool disposing)
		{
			if (disposing) {
				// See if either of the bitmaps used need clean-up.  
				if (desktop != null) desktop.Dispose();
				if (designModeDesktop != null) designModeDesktop.Dispose();
			}
			base.Dispose(disposing);
		}

		protected override void OnPaint(PaintEventArgs pe)
		{
			// If the control is in design mode, draw a nice background, otherwise paint the desktop.
			if (!DesignMode) {
				System.Diagnostics.Debug.Assert(desktop != null);
				DrawDesktopImage(desktop, pe.Graphics);
            } else {
				// Draw a static screenshot of a Windows desktop to simulate the control in action
				System.Diagnostics.Debug.Assert(designModeDesktop != null);
				DrawDesktopImage(designModeDesktop, pe.Graphics);
			}
			base.OnPaint(pe);
		}

		/// <summary>
		/// Draws an image onto the control in a size-aware way.
		/// </summary>
		/// <param name="desktopImage">The desktop image to be drawn to the control's sufrace.</param>
		/// <param name="g">The Graphics object representing the control's drawable surface.</param>
		protected void DrawDesktopImage(Image desktopImage, Graphics g)
		{
			// See if the image needs to be clipped (i.e., it is too big for the 
			// available space) or centered (i.e., it is too small)
			int x, y;
			
			if (ClientSize.Width > desktopImage.Size.Width) {
				x = (ClientRectangle.Width - desktopImage.Width) / 2;
			} else {
				x = DisplayRectangle.X;
			}

			if (ClientSize.Height > desktopImage.Size.Height ) {
				y = (ClientRectangle.Height- desktopImage.Height) / 2;
			} else {
				y = DisplayRectangle.Y;
			}

			g.DrawImage(desktopImage, x, y);
		}


		/// <summary>
		/// Dispatches the ConnectionLost event if any targets have registered.
		/// </summary>
		/// <param name="e">An EventArgs object.</param>
		protected void OnConnectionLost(EventArgs e)
		{
			if (ConnectionLost != null) {
				ConnectionLost(this, e);
			}
		}
		
		/// <summary>
		/// Dispatches the ConnectComplete event if any targets have registered.
		/// </summary>
		/// <param name="e">A ConnectEventArgs object with information about the remote framebuffer's geometry.</param>
		protected void OnConnectComplete(ConnectEventArgs e)
		{
			if (ConnectComplete != null) {
				ConnectComplete(this, e);
			}
		}

		// Handle Mouse Events:		 -------------------------------------------
		// In all cases, the same thing needs to happen: figure out where the cursor
		// is, and then figure out the state of all mouse buttons.
		// TODO: currently we don't handle the case of 3-button emulation with 2-buttons.

		protected override void OnMouseMove(MouseEventArgs mea)
		{
			// Since there might be dead space surrounding the desktop, only bother sending
			// the update if the mouse is really over the image
			Rectangle desktopRect = vnc.Framebuffer.Rectangle;

			if (ClientSize.Width > desktop.Size.Width) {
				desktopRect.X = (ClientRectangle.Width - desktop.Width) / 2;
			}

			if (ClientSize.Height > desktop.Size.Height ) {
				desktopRect.Y = (ClientRectangle.Height - desktop.Height) / 2;
			}

			// See if the mouse pointer is inside the area occupied by the desktop on screen.
			if (desktopRect.Contains(PointToClient(MousePosition))) {
				UpdateRemotePointer();
			}

			base.OnMouseMove(mea);
		}

		protected override void OnMouseDown(MouseEventArgs mea)
		{
			UpdateRemotePointer();
			base.OnMouseDown(mea);
		}
		
		// Find out the proper masks for Mouse Button Up Events
		protected override void OnMouseUp(MouseEventArgs mea)
		{
			UpdateRemotePointer();
			base.OnMouseUp(mea);
		}
		
		// TODO: Perhaps overload UpdateRemotePointer to take a flag indicating if mousescroll has occured??
		protected override void OnMouseWheel(MouseEventArgs mea)
		{
			// HACK: this check insures that while in DesignMode, no messages are sent to a VNC Host
			// (i.e., there won't be one--NullReferenceException)			
			if (!DesignMode) {
				Point current = PointToClient(MousePosition);
				byte mask = 0;

				// mouse was scrolled forward
				if (mea.Delta > 0) {
					mask += 8;
				} else if (mea.Delta < 0) { // mouse was scrolled backwards
					mask += 16;
				}

				System.Diagnostics.Debug.Assert(vnc != null);
				vnc.WritePointerEvent(mask, new Point(current.X, current.Y));
			}			
			base.OnMouseWheel(mea);
		}
		
		private void UpdateRemotePointer()
		{
			// HACK: this check insures that while in DesignMode, no messages are sent to a VNC Host
			// (i.e., there won't be one--NullReferenceException)			
			if (!DesignMode) {
				Point current = PointToClient(MousePosition);
				byte mask = 0;

				if (Control.MouseButtons == MouseButtons.Left)   mask += 1;
				if (Control.MouseButtons == MouseButtons.Middle) mask += 2;
				if (Control.MouseButtons == MouseButtons.Right)  mask += 4;

				// Adjust the mouse location relative to the position of the desktop,
				// which could be clipped and scrolled, or centered in a larger window.
				Point adjusted = new Point();
				
				if (ClientSize.Width > desktop.Size.Width) {
					int dx = (ClientRectangle.Width - desktop.Width) / 2;
					adjusted.X = current.X - dx;
				} else {
					adjusted.X = current.X - AutoScrollPosition.X;
				}

				if (ClientSize.Height > desktop.Size.Height ) {
					int dy = (ClientRectangle.Height - desktop.Height) / 2;
					adjusted.Y = current.Y - dy;
				} else {
					adjusted.Y = current.Y - AutoScrollPosition.Y;
				}

				System.Diagnostics.Debug.Assert(vnc != null);
				vnc.WritePointerEvent(mask, adjusted);
			}
		}

		// Handle Keyboard Events:		 -------------------------------------------
		// TODO: currently we don't handle case of "helper" keys properly (e.g., Shift, Alt, etc.)

		// These keys don't normally throw an OnKeyDown event. Returning true here fixes this.
		protected override bool IsInputKey(Keys keyData )
		{
			bool IsInpKey = true;
			
			switch (keyData) {
				case Keys.Tab:
					break;
				case Keys.Up:
					break;
				case Keys.Down:
					break;
				case Keys.Left:
					break;
				case Keys.Right:
					break;
// TESTING: Here is the original code before trying to fix the "Shift" problem
//				case Keys.Up | Keys.Shift:
//					break;
//				case Keys.Down | Keys.Shift:
//					break;
//				case Keys.Left | Keys.Shift:
//					break;
//				case Keys.Right | Keys.Shift:
//					break;
				case Keys.Shift:
					break;
				default:
					IsInpKey = base.IsInputKey(keyData);
					break;
			}
			return IsInpKey;
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			// HACK: this check insures that while in DesignMode, no messages are sent to a VNC Host
			// (i.e., there won't be one--NullReferenceException)			
			if (!DesignMode) {
				if (e.Shift)	vnc.WriteKeyboardEvent(0xffe1, true);
				if (e.Alt)		vnc.WriteKeyboardEvent(0xffe9, true);
				if (e.Control)	vnc.WriteKeyboardEvent(0xffe3, true);
//				if (e.KeyCode == Keys.Left){
//					// FIXME: Matt, can we remove this now???
//					Console.WriteLine("Keys.Up Pressed");
//				}

				// Get keycode
				System.Diagnostics.Debug.Assert(vnc != null);
				vnc.WriteKeyboardEvent(GetKey(e), true);
			}			
			base.OnKeyDown(e);
		}

		protected override void OnKeyUp(KeyEventArgs e)
		{
			// HACK: this check insures that while in DesignMode, no messages are sent to a VNC Host
			// (i.e., there won't be one--NullReferenceException)			
			if (!DesignMode) {
				if (!e.Shift)	vnc.WriteKeyboardEvent(0xffe1, false);
				if (!e.Alt)		vnc.WriteKeyboardEvent(0xffe9, false);
				if (!e.Control)	vnc.WriteKeyboardEvent(0xffe3, false);

				// Get keycode
				System.Diagnostics.Debug.Assert(vnc != null);
				vnc.WriteKeyboardEvent(GetKey(e), false);
			}				
			base.OnKeyUp(e);
		}

		// TODO: this needs to be moved down to a protocol level class...
		private uint GetKey(KeyEventArgs e)
		{
			uint keycode = (uint)e.KeyCode;
			switch(e.KeyCode) {
				case Keys.Back:
					keycode = 0xff08;
					break;
				case Keys.Tab:
					keycode = 0xff09;
					break;
				case Keys.Return:
					keycode = 0xff0d;
					break;
				case Keys.Escape:
					keycode = 0xff1b;
					break;
				case Keys.Insert:
					keycode = 0xff63;
					break;
				case Keys.Delete:
					keycode = 0xffff;
					break;
				case Keys.Home:
					keycode = 0xff50;
					break;
				case Keys.End:
					keycode = 0xff57;
					break;
				case Keys.PageUp:
					keycode = 0xff55;
					break;
				case Keys.PageDown:
					keycode = 0xff56;
					break;
				case Keys.Left:
					keycode = 0xff51;
					break;
				case Keys.Up:
					keycode = 0xff52;
					break;
				case Keys.Right:
					keycode = 0xff53;
					break;
				case Keys.Down:
					keycode = 0xff54;
					break;
				case Keys.F1:
					keycode = 0xffbe;
					break;
				case Keys.F2:
					keycode = 0xffbf;
					break;
				case Keys.F3:
					keycode = 0xffc0;
					break;
				case Keys.F4:
					keycode = 0xffc1;
					break;
				case Keys.F5:
					keycode = 0xffc2;
					break;
				case Keys.F6:
					keycode = 0xffc3;
					break;
				case Keys.F7:
					keycode = 0xffc4;
					break;
				case Keys.F8:
					keycode = 0xffc5;
					break;
				case Keys.F9:
					keycode = 0xffc6;
					break;
				case Keys.F10:
					keycode = 0xffc7;
					break;
				case Keys.F11:
					keycode = 0xffc8;
					break;
				case Keys.F12:
					keycode = 0xffc9;
					break;

//    All taken care of other places
//				case Keys.LShiftKey:
//					keycode = 0xffe1;
//					break;
//				case Keys.RShiftKey:
//					keycode = 0xffe2;
//					break;
//				case Keys.LControlKey:
//					keycode = 0xffe3;
//					break;
//				case Keys.RControlKey:
//					keycode = 0xffe4;
//					break;
//				case Keys.Alt:
//					keycode = 0xffe9;
//					break;
				default:
					break;
			}

			// split these into seperate if statements just 
			// to know which keys are being manipulated
			if (keycode > 64 && keycode <= 90) {
				// A-Z
				if (!e.Shift) keycode += 32;
			} else if (keycode > 47 && keycode <= 57) { 
				//0-9 (well...1-0)
				if (e.Shift) {
					switch (keycode) {
						case 48:  // 0
							keycode -= 7;
							break;
						case 50:  // 2
							keycode += 14;
							break;
						case 49:  // 1
						case 51:  // 3
						case 52:  // 4
						case 53:  // 5
							keycode -= 16;
							break;
						case 54:  // 6
							keycode += 40;
							break;
						case 55:  // 7
						case 57:  // 9
							keycode -= 17;
							break;
						case 56:  // 8
							keycode -= 14;
							break;
					}
				}
			} else {
				switch (keycode) {
					case 192: // `
						keycode = 96;
						if (e.Shift) keycode += 30; // 126;
						break;
					case 188: // ,
						keycode = 44;
						if (e.Shift) keycode += 16;
						break;
					case 190: // .
						keycode = 46;
						if (e.Shift) keycode += 16;
						break;
					case 191: // /
						keycode = 47;
						if (e.Shift) keycode += 16;
						break;
					case 186: // ;
						keycode = 59;
						if (e.Shift) keycode -= 1;
						break;
					case 222: // '
						keycode = 39;
						if (e.Shift) keycode -= 5;
						break;
					case 219: // [
						keycode = 91;
						if (e.Shift) keycode += 32;
						break;
					case 221: // ]
						keycode = 93;
						if (e.Shift) keycode += 32;
						break;
					case 189: // -
						keycode = 45;
						if (e.Shift) keycode += 50;
						break;
					case 187: // =
						keycode = 61;
						if (e.Shift) keycode -= 18;
						break;
					case 220: // \
						keycode = 92;
						if (e.Shift) keycode += 32;
						break;
					default:
						break;
				}
			}
			return keycode;
		}
	}
}