// VncSharp - .NET VNC Client Library
// Copyright (C) 2004  David Humphrey, Chuck Borgh, Matt Cyr
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

using System;
using System.Threading;
using System.Drawing;
using System.Diagnostics;
using System.Security.Cryptography;
using System.Windows.Forms;

using VncSharp.Encodings;

namespace VncSharp
{
	/// <summary>
	/// Delegate definition of an Event Handler used to indicate a Framebuffer Update has been received.
	/// </summary>
	public delegate void VncUpdateHandler(object sender, VncEventArgs e);

	public class VncClient
	{
		RfbProtocol rfb;			// The protocol object handling all communication with server.
		Framebuffer buffer;			// The geometry and properties of the remote framebuffer
		byte securityType;			// The type of Security agreed upon by client/server
		EncodedRectangleFactory factory;
		Thread worker;				// To request and read in-coming updates from server
		ManualResetEvent done;		// Used to tell the worker thread to die cleanly
	
		public VncClient()
		{
			rfb = new RfbProtocol();
		}

		/// <summary>
		/// Gets the Framebuffer representing the remote server's desktop geometry.
		/// </summary>
		public Framebuffer Framebuffer {
			get { 
				return buffer; 
			}
		}

		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <param name="display">The Display number (used on Unix hosts).</param>
		/// <param name="port">The Port number used by the Host, usually 5900.</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host, int display, int port)
		{
			if (host == null) throw new ArgumentNullException("host");

			// If a diplay number is specified (used to connect to Unix servers)
			// it must be 0 or greater.  This gets added to the default port number
			// in order to determine where the server will be listening for connections.
			if (display < 0) throw new ArgumentOutOfRangeException("display", display, "Display number must be non-negative.");
			port += display;
			
			// Connect and determine version of server, and set client protocol version to match			
			try {
				rfb.Connect(host, port);
				rfb.ReadProtocolVersion();
				rfb.WriteProtocolVersion();

				// Figure out which type of authentication the server uses
				byte[] types = rfb.ReadSecurityTypes();
				
				// Based on what the server sends back in the way of supported Security Types, one of
				// two things will need to be done: either the server will reject the connection (i.e., type = 0),
				// or a list of supported types will be sent, of which we need to choose and use one.
				if (types.Length > 0) {
					if (types[0] == 0) {
						// The server is not able (or willing) to accept the connection.
						// A message follows indicating why the connection was dropped.
						throw new VncProtocolException("Connection Failed. The server rejected the connection for the following reason: " + rfb.ReadSecurityFailureReason());
					} else {
						securityType = GetSupportedSecurityType(types);
						Debug.Assert(securityType > 0, "Unknown Security Type(s)", "The server sent one or more unknown Security Types.");
						
						rfb.WriteSecurityType(securityType);
						
						// Protocol 3.8 states that a SecurityResult is still sent when using NONE (see 6.2.1)
						if (rfb.ServerVersion == 3.8f && securityType == 0) {
							if (rfb.ReadSecurityResult() > 0) {
								// For some reason, the server is not accepting the connection.  Get the
								// reason and throw and exception
								throw new VncProtocolException("Unable to Connecto to the Server. The Server rejected the connection for the following reason: " + rfb.ReadSecurityFailureReason());
							}
						}
						
						return (securityType > 1) ? true : false;
					}
				} else {
					// Something is wrong, since we should have gotten at least 1 Security Type
					throw new VncProtocolException("Protocol Error Connecting to Server. The Server didn't send any Security Types during the initial handshake.");
				}
			} catch (Exception e) {
				throw new VncProtocolException("Unable to connect to the server. Error was: " + e.Message);
			}			
		}

		#region	Connect() method Overloads
		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required. Default Display and Port numbers are used.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host)
		{
			return Connect(host, 0, 5900);
		}

		/// <summary>
		/// Connect to a VNC Host and determine which type of Authentication it uses. If the host uses Password Authentication, a call to Authenticate() will be required. The Port number is calculated based on the Display.
		/// </summary>
		/// <param name="host">The IP Address or Host Name of the VNC Host.</param>
		/// <param name="display">The Display number (used on Unix hosts).</param>
		/// <returns>Returns True if the VNC Host requires a Password to be sent after Connect() is called, otherwise False.</returns>
		public bool Connect(string host, int display)
		{
			return Connect(host, display, 5900);
		}
		#endregion		

		/// <summary>
		/// Examines a list of Security Types supported by a VNC Server and chooses one that the Client supports.  See 6.1.2 of the RFB Protocol document v. 3.8.
		/// </summary>
		/// <param name="types">An array of bytes representing the Security Types supported by the VNC Server.</param>
		/// <returns>A byte that represents the Security Type to be used by the Client.</returns>
		protected byte GetSupportedSecurityType(byte[] types)
		{
			// Pick the first match in the list of given types.  If you want to add support for new
			// security types, do it here:
			for (int i = 0; i < types.Length; ++i) {
				if (   types[i] == 1  	// None
					|| types[i] == 2	// VNC Authentication
// None of the following are currently supported -------------------
//					|| types[i] == 5	// RA2
//					|| types[i] == 6    // RA2ne
//					|| types[i] == 16   // Tight
//					|| types[i] == 17 	// Ultra
//					|| types[i] == 18 	// TLS
				   ) return types[i];
			}
			return 0;
		}

		/// <summary>
		/// Use a password to authenticate with a VNC Host. NOTE: This is only necessary if Connect() returns TRUE.
		/// </summary>
		/// <param name="password">The password to use.</param>
		/// <returns>Returns True if Authentication worked, otherwise False.</returns>
		public bool Authenticate(string password)
		{
			if (password == null) throw new ArgumentNullException("password");
			
			// If new Security Types are supported in future, add the code here.  For now, only 
			// VNC Authentication is supported.
			if (securityType == 2) {
				PerformVncAuthentication(password);
			} else {
				throw new NotSupportedException("Unable to Authenticate with Server. The Server uses an Authentication scheme unknown to the client.");
			}
			
			if (rfb.ReadSecurityResult() == 0) {
				return true;
			} else {
				// Authentication failed, and if the server is using Protocol version 3.8, a 
				// plain text message follows indicating why the error happend.  I'm not 
				// currently using this message, but it is read here to clean out the stream.
				// In earlier versions of the protocol, the server will just drop the connection.
				if (rfb.ServerVersion == 3.8) rfb.ReadSecurityFailureReason();
				rfb.Close();	// TODO: Is this the right place for this???
				return false;
			}
		}

		/// <summary>
		/// Performs VNC Authentication using VNC DES encryption.  See the RFB Protocol doc 6.2.2.
		/// </summary>
		/// <param name="password">A string containing the user's password in clear text format.</param>
		protected void PerformVncAuthentication(string password)
		{
			byte[] challenge = rfb.ReadSecurityChallenge();
			rfb.WriteSecurityResponse(EncryptChallenge(password, challenge));
		}

		/// <summary>
		/// Encrypts a challenge using the specified password. See RFB Protocol Document v. 3.8 section 6.2.2.
		/// </summary>
		/// <param name="password">The user's password.</param>
		/// <param name="challenge">The challenge sent by the server.</param>
		/// <returns>Returns the encrypted challenge.</returns>
		protected byte[] EncryptChallenge(string password, byte[] challenge)
		{
			byte[] key = new byte[8];

			// Key limited to 8 bytes max.
			if (password.Length >= 8) {
				System.Text.Encoding.ASCII.GetBytes(password, 0, 8, key, 0);
			} else {
				System.Text.Encoding.ASCII.GetBytes(password, 0, password.Length, key, 0);
			}			

			// HACK: VNC reverses key order!?
			key = FixBug(key); 

			// VNC uses DES, not 3DES as written in some documentation
			DES des = new DESCryptoServiceProvider();
			des.FeedbackSize = 0;
			des.Padding = PaddingMode.None;
			des.Mode = CipherMode.ECB;

			ICryptoTransform enc = des.CreateEncryptor(key, null); 

			byte[] response = new byte[16];
			enc.TransformBlock(challenge, 0, challenge.Length, response, 0);
			
			return response;
		}

		/// <summary>VNC DES authentication has a bug, such that keys are reversed.  This code 
		/// was written by Dominic Ullmann (dominic_ullmann@swissonline.ch) and is 
		/// is being used under the GPL.</summary>
		/// <param name="desKey">The key to be altered.</param>
		/// <returns>Returns the fixed key as an array of bytes.</returns>
		protected byte[] FixBug(byte[] desKey)
		{
			byte[] newkey = new byte[8];

			for (int i = 0; i < 8; i++) {
				// revert desKey[i]:
				newkey[i] = (byte) (	
					((desKey[i] & 0x01) << 7) |  
					((desKey[i] & 0x02) << 5) |
					((desKey[i] & 0x04) << 3) |
					((desKey[i] & 0x08) << 1) |
					((desKey[i] & 0x10) >> 1) |
					((desKey[i] & 0x20) >> 3) |
					((desKey[i] & 0x40) >> 5) |
					((desKey[i] & 0x80) >> 7)
					);
			}
			return newkey;
		}
		
		/// <summary>
		/// Finish setting-up protocol with VNC Host.  Should be called after Connect and Authenticate (if password required).
		/// </summary>
		public void Initialize()
		{
			// Finish initializing protocol with host
			rfb.WriteClientInitialisation(false);
			buffer = rfb.ReadServerInit();
			rfb.WriteSetPixelFormat(buffer);	// just use the server's framebuffer format

			// TODO: currently Zrle is NOT supported...., 16});
			rfb.WriteSetEncodings(new uint[] {5, 4, 2, 1, 0}); 
			
			// Create an EncodedRectangleFactory so that EncodedRectangles can be built according to set pixel layout
			factory = new EncodedRectangleFactory(rfb, buffer);
		}

		/// <summary>
		/// Begin getting updates from the VNC Server.  This will continue until StopUpdates() is called.  NOTE: this must be called after Connect() and Authenticate() (if authentication is necessary).
		/// </summary>
		public void StartUpdates()
		{
			// Start getting updates on background thread.
			worker = new Thread(new ThreadStart(this.GetRfbUpdates));
			worker.IsBackground = true;
			done = new ManualResetEvent(false);
			
//			// Just for debugging!
//			worker.Name = "GetRfbUpdates_Worker";
//			Thread.CurrentThread.Name = "Main_Thread";

			worker.Start();
		}

		public void StopUpdates()
		{
			// FIXME: This isn't working yet.  The worker doesn't seem to die, despite setting the event.
			// Perhaps it is locked in a synchronous call to Draw via the Update event???
			done.Set();
			worker.Join(2000);
			rfb.Close();	
		}

		/// <summary>
		/// An event that occurs whenever the server sends a Framebuffer Update.
		/// </summary>
		public event VncUpdateHandler VncUpdate;
		
		/// <summary>
		/// Worker thread lives here and processes protocol messages infinitely, triggering events or other actions as necessary.
		/// </summary>
		private void GetRfbUpdates()
		{
			int rectangles;
			int enc;

			while (!done.WaitOne(0, false)) {
				rfb.WriteFramebufferUpdateRequest(0, 0, (ushort) buffer.Width, (ushort) buffer.Height, true);
				
				switch (rfb.ReadServerMessageType()) {
					case RfbProtocol.FRAMEBUFFER_UPDATE:
						rectangles = rfb.ReadFramebufferUpdate();

						for (int i = 0; i < rectangles; ++i) {
							// Get the update rectangle's info
							Rectangle rectangle;
							rfb.ReadFramebufferUpdateRectHeader(out rectangle, out enc);

							// Build a derived EncodedRectangle type and pull-down all the pixel info
							EncodedRectangle er = factory.Build(rectangle, enc);
							er.Decode();

							// Let the UI know that an updated rectangle is available, but check
							// to see if the user closed things down first.
							if (!done.WaitOne(0, false) & VncUpdate != null) {
								VncEventArgs e = new VncEventArgs(er);

								// In order to play nicely with WinForms controls, we do a check here to 
								// see if it is necessary to synchronize this event with the UI thread.
								if (VncUpdate.Target is System.Windows.Forms.Control) {
									Control target = VncUpdate.Target as Control;
									target.Invoke(VncUpdate, new object[] {this, e});
								} else {
									// Target is not a WinForms control, so do it on this thread...
									VncUpdate(this, new VncEventArgs(er));
								}
							}
						}
						break;
					case RfbProtocol.BELL:
						Beep(500, 300);  // TODO: are there better values than these?
						break;
					case RfbProtocol.SERVER_CUT_TEXT:
						Clipboard.SetDataObject(rfb.ReadServerCutText().Replace("\n", Environment.NewLine), true);
						break;
					case RfbProtocol.SET_COLOUR_MAP_ENTRIES:
// TODO: Needs to be implemented fully
//						rfb.ReadColourMapEntry();
						break;
				}
			}
		}

// There is no managed way to get a system beep (until Framework v.2.0). So depending on the platform, something external has to be called.
#if Win32
		[System.Runtime.InteropServices.DllImport("kernel32.dll")]
		private static extern bool Beep(int freq, int duration);
#else
		private bool Beep(int freq, int duration)	// bool just so it matches the Win32 API signature
		{
			// TODO: How to do this under Unix?
			System.Console.Write("Beep!");
			return true;
		}
#endif

		// TODO: This needs to be pushed into the protocol rather than expecting keysym from the caller.
		public void WriteKeyboardEvent(uint keysym, bool pressed)
		{
			rfb.WriteKeyEvent(keysym, pressed);
		}

		// TODO: This needs to be pushed into the protocol rather than expecting the caller to create the mask.
		public void WritePointerEvent(byte buttonMask, Point point)
		{
			rfb.WritePointerEvent(buttonMask, point);
		}
	}
}